from pyquil import Program
from pyquil.api import WavefunctionSimulator, QVMConnection
from pyquil.gates import *
from netQuil import *

__all__ = ["cat_entangler", "cat_disentangler", "QFT"]

def QFT(program, register):
    '''
    Performs Quantum Fourier Transform.

    :param Program program: program for where to apply QFT
    :param List register: register of qubits to perform QFT on
    :return: program where QFT has been applied
    '''
       
    n = len(register)
    for i in range(n-1, -1, -1):
        program += H(register[i])
        for j in range(i-1, -1, -1):
            k = i-j+1
            program += CPHASE(2*np.pi/(2**k),register[j], register[i])

    return program
    
def distributed_gate(agents):
    for agent in agents: agent.using_distributed_gate = not agent.using_distributed_gate

def notify_entangler_is_done(caller, target_agents): 
    ''' 
    Target agents should be expecting a cbit signaling the distributed gate is complete
    '''
    for agent in target_agents:
        cbit = [1] # cbit signaling done
        caller.csend(agent.name, cbit)

def cat_entangler(control, targets, entangled=False, notify=False):
    '''
    Performs the cat entangler, one of two primitive operations for 
    distributed quantum computing which can be used to implement non-local operations.
    Projects the state of Alice's local control bit (psi) on entangled qubits owned by other Agents, 
    allowing Agents to effectively use Alice's qubit as a control for their own operations. Measurements
    are stored in ro in the position corresponding to qubit's index (i.e. ith qubit measured into ro[i]). 
    Once the cat_entangler is started in the control agent, each target agent waits until they have received
    two classical bits from the control. The first is a placeholder for a measurement and the second indicates that
    the cat entangler is complete. e.g. non-local CNOTs, non-local controlled gates, and teleportation.

    Remember, the cat_disentangler sends classical bits between the control and targets, so be careful 
    when trying to perform similar operations in parallel!

    :param agent,int,int,register control: tuple of agent owning phi, phi, measurement qubit, and register for measurements
    :param List<agent,qubit> targets: list of tuples of agent and agent's qubit that will be altered
    :param Boolean entangled: true if qubits from other Agents are already maximally entangled
    :param Boolean notify: if true control agent will send cbit equaling 1 to all target agents, signaling completion 
    '''
    agent, psi, measure_qubit, ro = control
    p = agent.program

    # Collect all qubits except control bit
    qubits = [measure_qubit] + [q[1] for q in targets]

    # Tell Tracer to Ignore Operations
    distributed_gate([agent] + [t[0] for t in targets])

    # If qubits are not already entangled, and distribute all non-control qubits
    if not entangled:
        p += H(measure_qubit)
        for i in range(len(qubits) - 1):
            q1 = qubits[i] 
            q2 = qubits[i+1]
            p += CNOT(q1, q2)
    
    # Project control qubit onto qubit and measure
    p += CNOT(psi, measure_qubit)
    p += MEASURE(measure_qubit, ro[measure_qubit]) 

    # Send result of qubit that has been measured to all other agents
    cbit = [1] # Hard code a placeholder value until ro is calculated
    for a in [t[0] for t in targets]:
        agent.csend(a.name, cbit)

    # Conditionally Perform Measure Operations 
    for q in qubits:
        p.if_then(ro[measure_qubit], X(q))

    # Tell Tracer We're Done
    distributed_gate([agent] + [t[0] for t in targets])
    if notify: notify_entangler_is_done(caller=agent, target_agents=[t[0] for t in targets])

def cat_disentangler(control, targets, notify=False):
    '''
    Performs the cat disentangler, second of two primitive operations for 
    distributed quantum computing which can be used to implement non-local operations.
    Restores all qubits to state before cat_entangler was performed. Measurements
    are stored in ro in the position corresponding to qubit's index (i.e ith qubit measured into ro[i])

    Remember, the cat_disentangler sends classical bits between the control and targets, so be careful 
    when trying to perform similar operations in parallel!

    :param agent,int,register control: tuple of agent owning phi, phi, and register to store measurments
    :param List<agent,qubit> targets: list of tuples of agent and agent's qubit that will be altered
    :param Boolean notify: if true control agent will send cbit equaling 1 to all target agents, signaling completion 
    '''
    agent, psi, ro = control
    p = agent.program

    # Tell Tracer to Ignore Operations
    distributed_gate([agent] + [t[0] for t in targets])

    # Perform Hadamard of each qubit
    for _, q in targets: p += H(q)

    # Measure target bits and execute bit-flip to restore to 0
    for a, q in targets: 
        p += MEASURE(q, ro[q])
        p.if_then(ro[q], X(q))

        # Send qubit measurement to owner of psi in order to perform XOR
        cbit = [1] # Hard code a placeholder value until ro[q] is calculated
        a.csend(agent.name, cbit)
        agent.crecv(a.name)

    # Perform XOR between all measured bits and if true perform Z rotation on psi.
    # This will restore its original state
    for i in range(len(targets) - 1, 0, -1):
        p += XOR(ro[targets[i-1][1]], ro[targets[i][1]])

    p.if_then(ro[targets[0][1]], Z(psi))

    # Tell Tracer We're Done
    distributed_gate([agent] + [t[0] for t in targets])
    if notify: notify_entangler_is_done(caller=agent, target_agents=[t[0] for t in targets])

    


import inspect
import sys
import threading
import time

__all__ = ["Agent"]

class Agent(threading.Thread):
    def __init__(self, program=None, qubits=[], cmem=[], name=None):
        '''
        Agents are codified versions of Alice and Bob (i.e. single nodes in a quantum network) 
        that can send and receive classical and quantum information over connections. 
        Agents have the following properties: 

        * Their run-time logic is in the form of an Agent.run() method
        * Their connections to other agents are by default ingress and egress 
        * Agents' manage their own target and source devices for noise and local time tracking
        * Agents have a network monitor to record the traffic they see

        :param PyQuil<Program> program: program
        :param List<int> qubits: list of qubits owned by agent
        :param List<int> cmem: list of cbits owned by agent
        :param String name: name of agent, defaults to name of class
        '''
        threading.Thread.__init__(self)

        # Name of the agent, e.g. "Alice". Defaults to the name of the class.
        if name is not None:
            self.name = name
        else:
            self.name = self.__class__.__name__

        self.time = 0.0
        self.pulse_length = 10 * 10 ** -12 # 10 ps default photon pulse length
        self.qconnections = {}
        self.cconnections = {}

        # Define qubits, corresponding classical memory and program
        self.qubits = qubits 
        self.cmem = cmem 
        self.program = program

        # Define Agent Devices
        self.target_devices = []
        self.source_devices = []

        self.master_clock = None
        self.network_monitor_running = False
        self.using_distributed_gate = False

    def get_master_time(self): 
        '''
        :return: master time
        '''
        return self.master_clock.get_time()

    def _start_tracer(self):
        threading.settrace(self._tracer)

    def _get_device_results(self):
        '''
        Get results for source and target devices
        '''
        devices = self.source_devices + self.target_devices
        for device in devices: 
            if device.verbose:
                device.get_results()
 
    def update_network_monitor(self, qubits, bar):
        for _ in qubits: 
            time.sleep(0.05)
            bar.update()

    def _tracer(self, frame, event, arg):
        '''
        Prevents agent from modifying qubits that it does not own and manage by
        examining the frame and intercepting all pyquil.gates calls 
        '''
        if event == "call":
            if self.using_distributed_gate: 
                return self._tracer
            if frame.f_globals['__name__'] == 'pyquil.gates':
                # Returns dictionary of parameter names and their values
                argsToGate = inspect.getargvalues(frame)
                # Extract parameter values 
                argValues = list(argsToGate.locals.values())
                # Extract parameter values that are integers (ignore iterator arguments)
                qubits = [q for q in argValues if type(q) == int]
                # Check that qubits are a subset of the Agent's qubits
                if not all(q in self.qubits for q in qubits):
                    raise Exception('Agent cannot modify qubits they do not own (including qubits that have been lost)')

        return self._tracer
    
    def set_program(self, program):
        '''
        Set agent's program

        :param PyQuil<Program> program: agent's program
        '''
        self.program = program

    def add_target_devices(self, new_target_devices):
        '''
            Add target devices. Every qubit sent to agent will pass through these devices, 
            and, if set, the device's effect and time-delay will be applied

            :param List<Device> new_target_devices: list of target devices
        '''
        self.target_devices.extend(new_target_devices)

    def add_source_devices(self, new_source_devices):
        '''
            Add source devices. Every qubit sent by Agent will pass through these devices, 
            and, if set, the device's effect and time-delay will be applied 

            :param List<Device> new_source_devices: list of source devices
        '''
        self.source_devices.extend(new_source_devices)

    @property
    def cmem(self): 
        return self.__cmem

    @cmem.setter
    def cmem(self, cmem):
        '''
            Set classical memory as list of 0s and 1s

            :param List<int> cmem: classical memory
        '''
        if len(cmem) >= 0 and all(bit == 0 or bit == 1 for bit in cmem):
            self.__cmem = cmem
        else:  
            raise Exception('Classical bits must be in list and either 0 or 1')

    def add_cmem(self, cbits):
        '''
            Add more classical memory in the form of a list of 0s and 1s

            :param List<int> cbits: classical memory to extend
        '''
        self.__cmem.extend(cbits) 

    def add_device(self, device_type, device):
        ''' 
            Add a device to an agent.

            :param String device_type: may be either `source` or `target`
            :param Device device: instance of device added
        '''
        if device_type == 'source': 
            self.source_devices.append(device)
        elif device_type == 'target':
            self.target_devices.append(device)
        else: 
            raise Exception('Invalid device type (e.g. \'source\' or \'target\'')

    def __hash__(self):
        '''
        Agents are hashed by their (unique) names
        '''
        return hash(self.name)
    
    def __eq__(self, other):
        '''
        Agents are compared for equality by their names.
        '''
        return self.name == other.name

    def __ne__(self, other):
        '''
        Agents are compared for inequality by their names
        '''
        return not (self == other)

    def qsend(self, target, qubits):
        '''
        Send qubits from agent to target. Connection will place qubits on queue 
        for target to retrieve. 

        :param String target: name of destination for qubits 
        :param List<int> qubits: list of qubits to send to destination
        '''
        # Raise exception if agent sends qubits they do no have
        if not set(qubits).issubset(set(self.qubits)): 
            raise Exception('Agent cannot send qubits they do not have')
            
        connection = self.qconnections[target]
        source_delay = connection.put(self.name, target, qubits, self.time)
    
        # Removing qubits being sent
        self.qubits = list(set(self.qubits) - set(qubits))

        # Update Agent's Time
        self.time += source_delay

        # Update Master Clock
        self.master_clock.record_qtransaction(self.time, 'sent', self.name, target, qubits)

        # Update network monitor 
        if self.network_monitor_running: 
            self.update_network_monitor(qubits, self.pbar_sent)

    def qrecv(self, source):
        '''
        Agent receives qubits from source. Adds qubits to agent's list of qubits and
        add time delay. Return list of qubits
        
        :param String source: name of agent who sent qubits.
        :return: list of qubits sent from source
        '''
        connection = self.qconnections[source]
        qubits, delay, source_time = connection.get(self)
        self.time = max(source_time + delay, self.time) 

        # Update Master Clock
        self.master_clock.record_qtransaction(self.time, 'received', source, self.name, qubits)
        # Update network monitor 
        if self.network_monitor_running: 
            self.update_network_monitor(qubits, self.pbar_recv)

        return qubits
        
    def csend(self, target, cbits):
        '''
        Sends classical bits from agent to target.

        :param String target: name of target agent
        :param List<int> cbits: indices of cbits source is sending to target
        '''
        connection = self.cconnections[target]
        source_delay  = connection.put(target, cbits)
        scaled_source_delay = source_delay*len(cbits)
        self.time += scaled_source_delay
        
        #Update Master Clock
        self.master_clock.record_ctransaction(self.time, 'sent', self.name, target, cbits)
        
    def crecv(self, source):
        '''
        Get cbits from source. 
        
        :param String source: name of Agent where cbits originated from.
        :return: list of cbits sent from source
        '''
        connection = self.cconnections[source]
        cbits, delay = connection.get(self.name)
        self.time += delay

        #Update Master Clock
        self.master_clock.record_ctransaction(self.time, 'received', source, self.name, cbits)

        return cbits

    def run(self):
        '''Run-time logic for the Agent; this method should be overridden in child classes.'''
        pass



import queue
import multiprocessing
import itertools
import sys

__all__ = ["QConnect", "CConnect"]

pulse_length_default = 10 * 10 ** -12 # 10 ps photon pulse length
signal_speed = 2.998 * 10 ** 5 #speed of light in km/s
fiber_length_default = 0.0

class QConnect: 
    def __init__(self, *args, transit_devices=[]):
        '''
        This is the base class for a quantum connection between multiple agents. 

        :param agents \*args: list of agents to connect
        :param List<Devices> transit_devices: list of devices qubits travel through 
        '''
        agents = list(args)
        self.agents = {}
        self.source_devices = {}
        self.target_devices = {}
        self.transit_devices = {}

        '''
        Create queue to keep track of multiple requests. Name of queue is name of
        target agent.  
        '''
        self.queues = {}
        for agent in agents:
            self.agents.update({agent.name: agent})
            self.source_devices.update({agent.name: agent.source_devices})
            self.target_devices.update({agent.name: agent.target_devices})
            self.transit_devices.update({agent.name: transit_devices})
            self.queues.update({agent.name: queue.Queue()})

            for agentConnect in agents:
                if agentConnect != agent:
                    agent.qconnections[agentConnect.name] = self

    def put(self, source, target, qubits, source_time):
        ''' 
        Constructs full list of devices that each qubit must travel through. Sends the qubits
        through source devices. Places qubits and a list of transit and target 
        devices on the queue. Queue is keyed on the target agent's name.
        
        :param String source: name of agent where the qubits being sent originated
        :param String target: name of agent receiving qubits
        :param Array qubits: array of numbers corresponding to qubits the source is sending 
        :param Float source_time: time of source agent before sending qubits
        :returns: time qubits took to pass through source devices
        '''
        source_devices = self.source_devices[source]
        transit_devices = self.transit_devices[source]
        target_devices = self.target_devices[target]

        non_source_devices = {
            "transit": transit_devices,
            "target": target_devices,
        }

        program = self.agents[source].program
        source_delay = 0

        # Keep track of qubits remaining
        traveling_qubits = qubits

        if not source_devices:
            source_delay += pulse_length_default
        else:
            # Keep track of qubits lost by each device
            total_lost_qubits = []
            for device in source_devices:
                # If qubits are still remaining 
                if traveling_qubits:
                    res = device.apply(program, traveling_qubits)
                    if 'lost_qubits' in res.keys(): 
                        lost_qubits = res['lost_qubits']
                        # Remove lost qubits from traveling qubits
                        traveling_qubits = list(set(traveling_qubits) - set(lost_qubits))
                        # Add lost_qubits lost from current device to total_lost_qubits
                        total_lost_qubits += lost_qubits
                    if 'delay' in res.keys(): source_delay += res['delay']

                else: break

            # Invert lost qubits and add to traveling qubits
            for q in total_lost_qubits: 
                if q == 0: total_lost_qubits.append(float("-inf"))
                else: total_lost_qubits.append(-q)
            traveling_qubits += total_lost_qubits

        # Scale source delay time according to number of qubits sent
        scaled_source_delay = source_delay*len(qubits) 

        self.queues[target].put((traveling_qubits, non_source_devices, scaled_source_delay, source_time))
        return scaled_source_delay

    def get(self, agent): 
        '''
        Pops qubits off of the agent's queue. Sends qubit through transit and target devices,
        simulating a quantum network. Return an array of the qubits that have been altered, as well as
        the time it took the qubit to travel through the network. Some qubits may be lost during transmission. If lost,
        their value will switch to negative, or, in the case of 0, be set to -inf

        :param Agent agent: agent receiving the qubits 
        :returns: list of qubits, time to pass through transit and target devices, and the source agent's time
        '''
        traveling_qubits, devices, source_delay, source_time = self.queues[agent.name].get()

        agent.qubits = list(set(traveling_qubits + agent.qubits))

        program = self.agents[agent.name].program
       
        transit_devices = devices["transit"]
        target_devices = devices["target"]

        # Number of qubits before any are lost 
        num_travel_qubits = len(traveling_qubits)
        travel_delay = 0

        if not transit_devices:
            travel_delay += fiber_length_default/signal_speed
        if not target_devices:
            travel_delay += 0
        
        total_lost_qubits = [q for q in traveling_qubits if q < 0 or q == float("-inf")] 
        remaining_qubits = [q for q in traveling_qubits if q >= 0]

        for device in list(itertools.chain(transit_devices, target_devices)):
            # If qubits are remaining 
            if remaining_qubits: 
                res = device.apply(program, traveling_qubits)
                if 'lost_qubits' in res.keys(): 
                    lost_qubits = res['lost_qubits']
                    # Remove lost qubits from traveling qubits
                    remaining_qubits = list(set(remaining_qubits) - set(lost_qubits))
                    # Add lost_qubits lost from current device to total_lost_qubits
                    total_lost_qubits += lost_qubits
                if 'delay' in res.keys(): travel_delay += res['delay']
            else: break

        # Remove traveling_qubits
        agent.qubits = list(set(agent.qubits) - set(traveling_qubits))   

        lost_qubits_flipped = []
        for q in total_lost_qubits: 
            if q == 0: lost_qubits_flipped.append(float("-inf"))
            else:
                lost_qubits_flipped.append(-q)

        # Add inverted lost qubits to remaining qubits
        traveling_qubits = remaining_qubits + lost_qubits_flipped
        agent.qubits += traveling_qubits
        scaled_delay = travel_delay*num_travel_qubits + source_delay
        return traveling_qubits, scaled_delay, source_time

class CConnect: 
    def __init__(self, *args, length=0.0):
        '''
        This is the base class for a classical connection between multiple agents. 

        :param agents \*args: list of agents to connect
        :param Float length: distance between first and second agent
        '''
        agents = list(args)
        self.agents = {}

        '''
        Create queue to keep track of multiple requests. Name of queue is name of
        target agent.  
        '''
        self.queues = {}

        for agent in agents:
            self.agents.update({agent.name: agent})
            self.queues.update({agent.name: queue.Queue()})

            for agentConnect in agents:
                if agentConnect != agent:
                    agent.cconnections[agentConnect.name] = self

        self.length = length

    def put(self, target, cbits):
        ''' 
        Places cbits on queue keyed on the target Agent's name

        :param String target: name of recipient of program
        :param Array cbits: array of numbers corresponding to cbits agent is sending
        :returns: time for cbits to travel
        '''
        csource_delay = pulse_length_default * 8 * sys.getsizeof(cbits)
        self.queues[target].put((cbits, csource_delay))
        return csource_delay

    def get(self, agent): 
        ''' 
        Pops cbits off of the agent's queue and adds travel delay

        :param String agent: name of the agent receiving the cbits
        :returns: cbits from source and time they took to travel
        '''
        cbits, source_delay = self.queues[agent].get()
        travel_delay = self.length/signal_speed
        
        scaled_delay = travel_delay*len(cbits) + source_delay

        return cbits, scaled_delay


import numpy as np
import uuid

from pyquil.gates import *
from pyquil.quil import DefGate
from pyquil.noise import pauli_kraus_map

__all__ = ["bit_flip", "phase_flip", "depolarizing_noise", "measure","normal_unitary_rotation"]

ro = None

def kraus_op_bit_flip(prob: float):
    noisy_I = np.sqrt(1-prob) * np.asarray([[1, 0], [0, 1]])
    noisy_X = np.sqrt(prob) * np.asarray([[0, 1], [1, 0]])
    return [noisy_I, noisy_X]


def kraus_op_phase_flip(prob: float):
    noisy_I = np.sqrt(1-prob) * np.asarray([[1, 0], [0, 1]])
    noisy_Z = np.sqrt(prob) * np.asarray([[1, 0], [0, -1]])
    return [noisy_I, noisy_Z]


def kraus_op_depolarizing_channel(prob: float):
    noisy_I = np.sqrt(1-prob) * np.asarray([[1, 0], [0, 1]])
    noisy_X = np.sqrt(prob/3) * np.asarray([[0, 1], [1, 0]])
    noisy_Z = np.sqrt(prob/3) * np.asarray([[1, 0], [0, -1]])
    noisy_Y = np.sqrt(prob/3) * np.asarray([[0, 0-1.0j], [0+1.0j, 0]])
    return [noisy_I, noisy_X, noisy_Y, noisy_Z]

def random_unitary(n):
    # draw complex matrix from Ginibre ensemble
    z = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    # QR decompose this complex matrix
    q, r = np.linalg.qr(z)
    # make this decomposition unique
    d = np.diagonal(r)
    l = np.diag(d) / np.abs(d)
    return np.matmul(q, l)



def bit_flip(program, qubit, prob: float):
    '''
    Apply a bit flip with probability 

    :param Program program: program to apply noise to
    :param Integer qubit: qubit to apply noise to 
    :param Float prob: probability of apply noise 
    '''
    unique_id = uuid.uuid1().int

    flip_noisy_I_definition = DefGate("flipNOISE" + str(unique_id), random_unitary(2))
    program += flip_noisy_I_definition
    
    program.define_noisy_gate("flipNOISE" + str(unique_id), [qubit], kraus_op_bit_flip(prob))
    program += ("flipNOISE" + str(unique_id), qubit)

def phase_flip(program, qubit, prob: float):
    '''
    Apply a phase flip with probability

    :param Program program: program to apply noise to
    :param Integer qubit: qubit to apply noise to 
    :param Float prob: probability of apply noise 
    '''

    unique_id = uuid.uuid1().int

    phase_noisy_I_definition = DefGate("phaseNOISE" + str(unique_id), random_unitary(2))
    program += phase_noisy_I_definition
    
    program.define_noisy_gate("phaseNOISE" + str(unique_id), [qubit], kraus_op_phase_flip(prob))
    program += ("phaseNOISE" + str(unique_id), qubit)

def depolarizing_noise(program, qubit, prob: float):
    '''
    Apply depolarizing noise with probability

    :param Program program: program to apply noise to
    :param Integer qubit: qubit to apply noise to 
    :param Float prob: probability of apply noise 
    '''
    unique_id = uuid.uuid1().int

    dp_noisy_I_definition = DefGate("dpNOISE" + str(unique_id), random_unitary(2))
    program += dp_noisy_I_definition
    
    program.define_noisy_gate("dpNOISE" + str(unique_id), [qubit], kraus_op_depolarizing_channel(prob))
    program += ("dpNOISE" + str(unique_id), qubit)

def measure(program, qubit, prob: float, name):
    '''
    Measure the qubit with probability

    :param Program program: program to apply noise to
    :param Integer qubit: qubit to apply noise to 
    :param Float prob: probability of apply noise 
    :param String name: name of quil classical register to measure to.
    :returns: None if qubit is not measured and qubit if qubit is measured
    '''
    if np.random.rand()> prob:
        global ro

        devices_ro_exists = False
        ro_exists = False
        for inst in program.instructions: 
            try: 
                if inst.name == name:
                    devices_ro_exists = True
                elif inst.name == "ro":
                    ro_exists = True
            except: pass

        if not ro_exists:
            ro = program.declare("ro", 'BIT', 1)
        elif ro_exists and not devices_ro_exists:
            ro = program.declare(name, 'BIT', 1)

        program += MEASURE(qubit, ro)
        return qubit
    return None

def normal_unitary_rotation(program, qubit, prob:float, variance):
    '''
    Apply X and Z rotation with probability

    :param Program program: program to apply noise to
    :param Integer qubit: qubit to apply noise to 
    :param Float prob: probability of apply noise 
    :param Float variance: variance of rotation angle
    '''
    if np.random.rand() > prob:
        x_angle, z_angle = np.random.normal(0, variance, 2)
        program += RX(x_angle, qubit)
        program += RZ(z_angle, qubit)
    


import inspect
from .clock import *

from pyquil import Program

__all__ = ["Simulation"]

# Attributes to keep on newly instantiated agent when resetting agents
THREADING_ATTRS = ['_started', '_stderr', '_tstate_lock', '_initialized', '_indent', '_is_stopped', '_target']

class Simulation:
    def __init__(self, *args):
        '''
        Initialize the simulation
        '''
        self.agents = list(args)
        self.pbars = {}

    def _create_agent_copies(self):
        '''
        Creates a dictionary of attributes for each agent. Stores list of dictionaries on self.agent_copies.
        Used to reset agents to their original state before each trial. Execute this function once at beginning of run
        if client requests multiple trials
        '''
        self.agent_copies = []
        program = self.agents[0].program.copy()
        for agent in self.agents:
            # Get attributes
            attributes = inspect.getmembers(agent, lambda a:not(inspect.isroutine(a)))
            # Remove private funcs and pythonic attrs 
            attrs_tuples = [a for a in attributes if not(a[0].startswith('__') and a[0].endswith('__'))]
            attrs_dict = dict(attrs_tuples)
            attrs_clean = {k: v for k, v in attrs_dict.items() if k not in THREADING_ATTRS}
            attrs_clean["program"] = program
            self.agent_copies.append(attrs_clean)

    def _reset_agents(self, agent_classes):
        '''
        Creates exact duplicate of original agent given the class constructor and 
        a dictionary of agent attributes from _create_agent_copies

        :param List<Agent> agent_classes: list of agent classes 
        '''
        program_copy = None
        for indx, copy in enumerate(self.agent_copies): 
            # Create copy of agent
            new_agent = agent_classes[indx]()
            for k in copy:
                try: 
                    setattr(new_agent, k, copy[k])
                    if k == 'program' and program_copy == None:
                        program_copy = copy[k].copy()
                    if k == 'program': 
                        setattr(new_agent, k, program_copy)
                except: 
                    pass
            self.agents[indx] = new_agent

    def _reset_devices(self):
        '''
        Reset source devices for each agent after each trial. 
        '''
        for agent in self.agents: 
            for connection in agent.qconnections.values():
                connection.agents = {a.name: a for a in self.agents}
            for device in agent.source_devices:
                device.reset()

    def _add_program(self):
        '''
        Adds program to all agents if none set. If agents are not all sharing
        the same program, raise an exception.
        '''
        p = Program()
        set_program = self.agents[0].program

        for agent in self.agents: 
            if agent.program != set_program:
                raise Exception('All agents must share the same program')
            if agent.program == None: 
                agent.program = p

    def run(self, trials=1, agent_classes=[], network_monitor=False):
        '''
        Run the simulation

        :param Int trials: number of times to simulate program
        :param List<Agent> agent_classes: list of agent classes
        :param Boolean network_monitor: outputs each network transaction and device information
        :return: returns list of programs. One for each trial
        '''
        # If program is not set, add default
        self._add_program()
        
        programs = []
        running_trials = trials > 1 

        # If trials is greater than 1, create copies of each agent
        if running_trials: self._create_agent_copies()

        for _ in range(trials): 

            # Start master clock and network monitor
            master_clock = MasterClock()
            for agent in self.agents:
                agent.master_clock = master_clock

            # Start agents, tracer, and network monitor
            for idx, agent in enumerate(self.agents):
                agent._start_tracer()
                agent.start()

            # Wait for agents to finish 
            for agent in self.agents: 
                agent.join()

            if network_monitor: 
                agent._get_device_results()
                master_clock.display_transactions()

            # Record program generated from trial
            programs.append(self.agents[0].program)

            # Reset agents if multiple trials
            if running_trials:
                self._reset_agents(agent_classes)
                self._reset_devices()

        return programs


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="netQuil",
        version = "1.0.1",
    author="Zac Espinosa, Matt Radzihovsky",
    author_email="zespinosa97@gmail.com, mattradz@stanford.edu",
    description="A Distributed Quantum Network Simulator",
    license="MIT",
    long_description="NetQuil is an open-source Python framework built on pyQuil designed specifically for simulating quantum networks.",
    long_description_content_type="text/markdown",
    url="https://github.com/att-innovate/netQuil",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy",
        "pyquil"
    ],
)



import inspect

__all__ = ["MasterClock"]

class MasterClock:
    def __init__(self):
        self.transactions = []
        self.time = 0

    def record_qtransaction(self, time, event_type, source, target, qubits):
        '''
        Update master clock and record quantum transaction

        :param Float time: current time in seconds
        :param String event_type: either sent or received. Otherwise, raise exception
        :param String source: name of the agent sending qubits
        :param String target: name of the agent receiving qubits
        :param List<int> qubits: list of qubits being sent from source to target
        '''
        # Update master clock
        self.time = max(self.time, time)

        # Record Event
        if event_type == 'sent':
            transaction = 'Qubits {} sent from {} to {} at {}'.format(qubits, source, target, time)
        elif event_type == 'received':
            transaction = 'Qubits {} received by {} from {} at {}'.format(qubits, target, source, time)
        else: 
            raise Exception('Event type must be "sent" or "received"') 

        self.transactions.append(transaction)

    def record_ctransaction(self, time, event_type, source, target, cbits):
            '''
            Update master clock and record classical transaction

            :param Float time: current time in seconds
            :param String event_type: either sent or received. Otherwise, raise exception
            :param String source: name of the agent sending cbits
            :param String target: name of the agent receiving cbits
            :param List<int> cbits: list of cbits being sent from source to target
            '''
            # Update master clock
            self.time = max(self.time, time)

            # Record Event
            if event_type == 'sent':
                transaction = 'Bits {} sent from {} to {} at {}'.format(cbits, source, target, time)
            elif event_type == 'received':
                transaction = 'Bits {} received by {} from {} at {}'.format(cbits, target, source, time)
            else: 
                raise Exception('Event type must be "sent" or "received"') 

            self.transactions.append(transaction)

    def get_time(self):
        '''
        :return: master time
        '''
        return self.time
    
    def recent_transaction(self):
        '''
        Print the most recent transaction
        '''
        print(self.transactions[-1])

    def display_transactions(self): 
        '''
        Prints all transactions
        '''
        for transaction in self.transactions: 
            print(transaction)


import numpy as np
import uuid

from pyquil import Program
from pyquil.gates import *
from netQuil import noise

__all__ = ["Fiber", "Laser", "Device"]

signal_speed = 2.998 * 10 ** 5 #speed of light in km/s

class Device(): 
    '''
    Base class for all source and target devices
    '''
    def __init__(self): 
        self.name = 'Device'
        self.success = 0
        self.trials = 0
    
    def apply(self, program, qubits):
        '''
        This function should be overwritten by children classes, and 
        should include runtime code defining how child devices impact qubits

        :param Pyquil<Program> program: program to manipulate
        :param List<int> qubits: list of qubits passing through device
        '''
        return {delay: None, qubits: None}

    def get_results(self):
        '''
        Prints device information about trial to console. 
        '''
        pass
    
    def reset(self): 
        '''
        reset is called between trials and resets all properties.
        ''' 
        pass 

class Fiber(Device):
    def __init__(self, length=0.0, attenuation_coefficient = -0.16, apply_error=True):
        '''
        Simulation of fiber optics with given length and attenuation coefficient. 

        :param Float length: length of fiber optical cable in km
        :param Float attenuation_coefficient: coefficient determining likelihood of photon loss
        :param Boolean apply_error: True is device should apply error, otherwise, only returns time delay
        '''
        decibel_loss = length*attenuation_coefficient
        self.attenuation = 10 ** (decibel_loss / 10)
        self.apply_error = apply_error
        self.length = length
        
    def apply(self, program, qubits):
        '''
        Applies device's error and returns time that photon took to pass through simulated device

        :param Program program: program to be modified
        :param List<int> qubits: qubits being sent
        :return: time qubits took to travel through fiber
        '''
        lost_qubits = []
        for i, qubit in enumerate(qubits):
            if qubit < 0: continue

            if self.apply_error:
                q = noise.measure(program, qubit, self.attenuation, "Fiber")
                if q is not None: 
                    lost_qubits.append(q)
                
        delay = self.length/signal_speed

        return {
            'delay': delay, 
            'lost_qubits': lost_qubits
        }

class Laser(Device):
    def __init__(self, pulse_length=10 * 10 ** -12, expected_photons=1.0, rotation_prob_variance=1.0, wavelength=1550, apply_error=True):
        '''
        Simulation of laser at 1550nm wavelength. Laser produce photons according to poisson
        distribution, centered around expected_photons. 
        '''
        self.variance = rotation_prob_variance
        self.wavelength = wavelength
        self.photon_expectation = expected_photons
        self.pulse_length = pulse_length
        self.apply_error = apply_error
        self.name = 'Laser'
        self.success = 0
        self.trials = 0
        
    def apply(self, program, qubits):
        '''
        Applies laser effect to qubits
        :param Program program: global program
        :param List<int> qubits: list of qubits going through laser
        :return: time it took qubits to pass through device
        '''
        for qubit in qubits:
            if qubit < 0: continue

            if self.apply_error:
                numPhotons = np.random.poisson(lam=self.photon_expectation)
                self.trials += len(qubits)
                if numPhotons == self.photon_expectation: self.success += 1
                '''
                Rotation Noise
                noise.normal_unitary_rotation(program, qubit, 0.5, self.variance)
                '''
        delay = self.pulse_length
        return {
            'delay': delay
        }

    def get_results(self):
        try: 
            print('{} has a signal to noise ratio of {}/{}'.format(self.name, self.success, self.trials))
        except:
            pass

    def reset(self):  
        self.success = 0
        self.trials = 0 



# netQuil: A Distributed Quantum Network Simulator
# Copyright (c) 2019 AT&T, Zac Espinosa, Matthew Radzihovsky
name = 'netQuil' 
# Load all modules
from netQuil.agents import *
from netQuil.simulator import *
from netQuil.connections import *
from netQuil.devices import *
from netQuil.noise import *
from netQuil.clock import *
from netQuil.distributedGates import *